/******************************************************************************
 *                                                                            *
 * Hoplite (c) Frans Ekman 2003                                               *
 *                                                                            *
 *                                                                            *
 * This file contains the position, move and movelist structs and prototypes  *
 * for functions that generates legal moves, edits the position, edits a move,*
 * edits the movelist, etc.                                                   *
 *                                                                            *
 * A position is stored in a 2d char array (matrix) where different chars     *
 * stand for the pieces. All the standard letters are used. Uppercase letters *
 * are used for white and lowercase for black.                                *
 * There are also fields for whose turn to move, 4 flags for the rights to    *
 * castle, The cordinates for the kings and a flag isCheck, isCheck = TRUE    *
 * if the current position is in check. A flag knowIfCheck is also needed,    *
 * so that we only have to find out if it is check once. I also have 2 flags  *
 * to remeber if the kings have moved. This is so that the evaluate function  *
 * would understand that an uncastled king that has moved is bad. Otherwise   *
 * the computer would move the king manually to the nearest corner.           *
 * 2 integers n_white_pieces and n_black_pieces to remember how many pieces   *
 * each side has left. This is used to find out if there is a risk for        *
 * zugswang in the null-move search.
 *                                                                            *
 * I use zobrist keys to compare positions. A position struct therefore also  *
 * has a 64bit zobrist key which is updated when the position changes.        *
 * There is still a 64 bit allPieces bitboard which was used for this earlier.*
 * It's "commented away", but I have left it there because it might be useful *
 * later.                                                                     *
 *                                                                            *
 * The move structure has fields for its 4 cordinates, a field for the        *
 * promotion piece (NO_PROMOTION for nonpromotion moves), a capiture flag,    *
 * and a field for the moving piece.                                          *
 *                                                                            *
 * Most important movegeneration functions are:                               *
 *                                                                            *
 *    PMoveList generateMoves(PPosition position),                            *
 *                                                                            *
 * which generates all legal moves, and                                       *
 *                                                                            *
 *    PMoveList generateCaptures(PPosition position),                         *
 *                                                                            *
 * which generates only captures. This is used in the quiesence search or     *
 * cature search to improve the speed.                                        *
 *                                                                            *
 * The function:                                                              *
 *                                                                            *
 *    void evaluateCheck(PPosition position),                                 *
 *                                                                            *
 * which finds out if it's check only if knowIfCheck = FALSE. It must be      *
 * called ALLWAYS after a move is made!                                       *
 *                                                                            *
 * When a move is added to the list with add() function it will be tested. If *
 * The move results in a position where the side that moved is in check, the  *
 * move is illegal, and will not be added. The function shall only be used    *
 * for normal moves, not castles. Castle moves are checked elsewere, and then *
 * added with the addWithoutCheck() function.                                 *
 *                                                                            *
 * Threaths can be evaluated with the functions threath_xy_z functions.       *
 * threath_lu_w(pos, x, y) will find out if the square (x,y) is threatened    *
 * by a black piece from the 135 degree direction (l=left, u=up and w=white   *
 * is threatened). threath_other_w(...) and threath_other_b(...) checks       *
 * knight and pawn threaths.                                                  *
 *                                                                            *
 * The functions:                                                             *
 *                                                                            *
 *     int resultInIllegalPositionForWhite(PPosition position, PMove move)    *
 *                                                                            *
 *     int resultInIllegalPositionForBlack(PPosition position, PMove move)    *
 *                                                                            *
 * are used to check if a non-kingmove will result in a position where the    *
 * side that moved is in check. This function uses all the threath_ functions *
 * It calculates the changes in x and y directions and choses the right       *
 * threath function according to that. If for instance dx=5 and dy=1 no tests *
 * are made, because no piece could be attacking from that direction.         *
 *                                                                            *
 * Some functions:                                                            *
 *                                                                            *
 *      int insufficientMaterial(PPosition position);                         *
 *                                                                            *
 * to check if its a draw by insufficient material.                           *
 *                                                                            *
 *      void setStartPosition(PPosition position);                            *
 *                                                                            *
 * to setup the startposition                                                 *
 *                                                                            *
 *      int setPosition(PPosition position, char *fenString);                 *
 *                                                                            *
 * to set any position by giving the position as a FEN string.                *
 *                                                                            *
 * modified 7.3.2004                                                          *
 *                                                                            *
 *****************************************************************************/


#ifndef POSITION_H
#define POSITION_H

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "def.h"
#include "zobrist.h"

typedef unsigned long long TBitboard;

extern TBitboard mask[64];
extern TBitboard wq_castle;
extern TBitboard wk_castle;
extern TBitboard bq_castle;
extern TBitboard bk_castle;
extern TBitboard wq_castle_mask;
extern TBitboard wk_castle_mask;
extern TBitboard bq_castle_mask;
extern TBitboard bk_castle_mask;

typedef struct {
  char board[8][8]; // x, y
  //TBitboard allPieces;
  zobristkey key; 
  int whiteShortCastle; 
  int blackShortCastle; 
  int whiteLongCastle; 
  int blackLongCastle;
  int whitesMove;
  int enPeassant; // the column
  int whiteKingSquare_X;
  int whiteKingSquare_Y;
  int blackKingSquare_X;  
  int blackKingSquare_Y;
  int isCheck;
  int knowIfCheck; // TRUE if we know for sure if we are in check
  int movesSince; // half moves since pawn move or capiture
  int whiteHasCastled;
  int blackHasCastled;
  int n_white_pieces;
  int n_black_pieces;
} TPosition;

typedef TPosition *PPosition;

typedef struct {
  short int x1;
  short int y1;
  short int x2;
  short int y2;
  char promotion; // small letters
  short int capiture;
  char piece; // small letters
} TMove;

typedef TMove *PMove;

typedef struct {
  PMove moves[MAX_POSSIBLE_MOVES];
  short int firstFree;
} TMoveList;

typedef TMoveList *PMoveList;

void freeMoveList(PMoveList list);

PMove newMove(short int x1, short int y1, short int x2, short int y2, 
              char promotion, short int capiture, char piece);

PMove createMove(PPosition position, char *moveString);

void add(PMoveList list, PMove move, PPosition position);

void addWithoutCheck(PMoveList list, PMove move);

void generateMask(void);

int toSquareNumber(int x, int y);

int getX(int square);

int getY(int square);

int convertX(char letter);

int convertToColumn(int x);

int setPosition(PPosition position, char *fenString);

void setStartPosition(PPosition position);

void removeCastleRights(PPosition position);

void doMove(PPosition position, PMove move);

int threath_lu_w(PPosition position, int x, int y);
int threath_ru_w(PPosition position, int x, int y);
int threath_ld_w(PPosition position, int x, int y);
int threath_rd_w(PPosition position, int x, int y);
int threath_u_w(PPosition position, int x, int y);
int threath_d_w(PPosition position, int x, int y);
int threath_l_w(PPosition position, int x, int y);
int threath_r_w(PPosition position, int x, int y);
int threath_other_w(PPosition position, int x, int y);

int threath_lu_b(PPosition position, int x, int y);
int threath_ru_b(PPosition position, int x, int y);
int threath_ld_b(PPosition position, int x, int y);
int threath_rd_b(PPosition position, int x, int y);
int threath_u_b(PPosition position, int x, int y);
int threath_d_b(PPosition position, int x, int y);
int threath_l_b(PPosition position, int x, int y);
int threath_r_b(PPosition position, int x, int y);
int threath_other_b(PPosition position, int x, int y);

int isSquareAttackedByWhite(PPosition position, int x, int y);
int isSquareAttackedByBlack(PPosition position, int x, int y);

void evaluateCheck(PPosition position);

int resultInIllegalPositionForWhite(PPosition position, PMove move);

int resultInIllegalPositionForBlack(PPosition position, PMove move);

int isLegal(PPosition position, PMove move);

void addPawnMoves(PPosition position, PMoveList list, int x, int y);

void addKnightMoves(PPosition position, PMoveList list, int x, int y);

void addBishopMoves(PPosition position, PMoveList list, int x, int y, char c);

void addRookMoves(PPosition position, PMoveList list, int x, int y, char c);

void addQueenMoves(PPosition position, PMoveList list, int x, int y);

void addCastleMoves(PPosition position, PMoveList list);

PMoveList generateMoves(PPosition position);

void printMove(PMove move);

void p_move(PMove move, FILE *file); 

void printMoveList(PMoveList list);

void printPosition(PPosition position);

int equals(PPosition pos_a, PPosition pos_b);

int move_equals(PMove a, PMove b);

int insufficientMaterial(PPosition position); 


///////////////////////////////////////////////
// Capture move generation for quiesence search


void addPawnCaptures(PPosition position, PMoveList list, int x, int y);

void addKnightCaptures(PPosition position, PMoveList list, int x, int y);

void addBishopCaptures(PPosition position, PMoveList list, int x, 
                       int y, char c);

void addRookCaptures(PPosition position, PMoveList list, int x, int y, char c);

void addQueenCaptures(PPosition position, PMoveList list, int x, int y);

PMoveList generateCaptures(PPosition position);




#endif


