/******************************************************************************
 *                                                                            *
 *  Hoplite (c) Frans Ekman 2003                                              *
 *                                                                            *
 *  search.c contains implementations of search function and functions that   *
 *  search uses.                                                              *
 *                                                                            *
 *  modified 7.3.2004                                                         *
 *                                                                            *
 *****************************************************************************/

#include "search.h"

PPosition helpPosition;
int atWhatPlyAreWe;
int nodes;
double timeOfStart;
int searchInterrupted;
int playPVLine;
int knowMoveScores;
TMoveScoreList moveScoresAtFirstPly;
TLine bestLine;

/******************************************************************************
 * The alpha-beta negamax search function                                     *
 *****************************************************************************/
int search(PGame game, int depth, int alpha, int beta, TLine *pline, 
           double maxtime, int ply, int maxSearchDepth, int previousNullMove,
           int hasBeenNullMove)
{
  PMoveList list;
  TGame gameCopy;
  int value;
  int i;
  int newDepth;
  TLine line;
  PMove tmpMove;

  int knowIfCheck, isCheck, enPeassant; // for null move

  int bestScore = -BIG_SCORE - 5;
  TMove bestMove;
  int hashFlag = HASH_ALPHA;

  int nrOfSpecialMoves = 0;

  int amount_of_extenssions_done = 0;

  bestMove.promotion = 'F';

  nodes++;
/******************************************************************************
 * Check the time and quit searching if we have spent too much time.          *
 *****************************************************************************/
  if (dtime() - timeOfStart > maxtime)
    searchInterrupted = TRUE;
  if (searchInterrupted) {
    pline->cmove = 0;
    return 0;
  }
 
/******************************************************************************
 * Check for repetition                                                       *
 *****************************************************************************/
  if (repeated3(game)) {
    pline->cmove = 0;
    return 0; 
  }

/******************************************************************************
 * Check the hashtables                                                       *
 *****************************************************************************/
  if (ply > 0) {
    value = probeHash(depth, alpha, beta, game->currentPosition.key,&bestMove);
    if (value != HASH_UNKNOWN_VALUE) {
      playPVLine = FALSE;
      pline->cmove = 0;
      if (value >= MUST_BE_MATE)
        return (value - ply);
      return value;
    }
  }

/******************************************************************************
 * try a null move if we are not in check or the previous move was not a      *
 * nullmove and there is still enough depth and pieces on the board           *
 *****************************************************************************/
  if (game->currentPosition.whitesMove)
    i = game->currentPosition.n_white_pieces;
  else 
    i = game->currentPosition.n_black_pieces;
  if (!(game->currentPosition.isCheck) && !previousNullMove && depth>0 && 
      i > 5) {
    /* make nullmove */
    game->currentPosition.whitesMove = !(game->currentPosition.whitesMove);
    knowIfCheck = game->currentPosition.knowIfCheck;
    isCheck = game->currentPosition.isCheck;
    enPeassant = game->currentPosition.enPeassant;
    game->currentPosition.knowIfCheck = TRUE;
    game->currentPosition.isCheck = FALSE;
    game->currentPosition.enPeassant = NO_ENPEASSANT;
    switchTurnToMove(&(game->currentPosition.key));

    value = -search(game, (depth-3), -beta, (-beta+1), &line, maxtime, 
                    ply+2, maxSearchDepth, TRUE, TRUE);

    if (value >= beta) {
      pline->cmove = 0;
      return beta;
    }
    
    /* unmake nullmove */
    game->currentPosition.whitesMove = !(game->currentPosition.whitesMove);
    switchTurnToMove(&(game->currentPosition.key));
    game->currentPosition.knowIfCheck = knowIfCheck;
    game->currentPosition.isCheck = isCheck;
    game->currentPosition.enPeassant = enPeassant;
  }

/******************************************************************************
 * Extend the search if we are in check and deep enough                       *
 *****************************************************************************/
  if (ply>1 && game->currentPosition.isCheck) {
    depth++;
    amount_of_extenssions_done++;
  }

/******************************************************************************
 * If we have searched deep enough, call quiesenceSearch() and set the number *
 * of moves in principal variation (pline->cmove) to 0.                       *
 *****************************************************************************/
  if (depth <= 0 || ply > maxSearchDepth) {
    pline->cmove = 0;
    playPVLine = FALSE;
    value = quiesenceSearch(game, alpha, beta); 
    return value;
  }
  
/******************************************************************************
 * At the root of the "chesstree" moves are ordered so that those moves that  *
 * scored high in the previous iteration comes first. If this is the first    *
 * iteration, just sort like quiesenceSearch.                                 *
 *****************************************************************************/
  if (ply == 0 && knowMoveScores) {
    list = (PMoveList)malloc(sizeof(TMoveList));
    list->firstFree = 0;
    for (i=0; i<moveScoresAtFirstPly.firstFree; i++) {
      tmpMove = (PMove)malloc(sizeof(TMove));
      *tmpMove = moveScoresAtFirstPly.moveScore[i].move;
      addWithoutCheck(list, tmpMove);
    }
  } else {
    list = generateMoves(&(game->currentPosition));
    if (list->firstFree > 1) {
      helpPosition = &(game->currentPosition);
      qsort(list->moves, list->firstFree-1, sizeof(PMove), fastcmpMoves);
    }
  }

  /* Extend search if there is only 1 legal move */
  if (list->firstFree == 1) {
    depth++;
    amount_of_extenssions_done++;
  }
  
/******************************************************************************
 * Check for mate or stalemate.                                               *
 *****************************************************************************/
  if (list->firstFree == 0) {
    if (game->currentPosition.isCheck) {
      freeMoveList(list);
      pline->cmove = 0;
      return -BIG_SCORE + ply;
    } else {
      pline->cmove = 0;
      freeMoveList(list);
      return 0;
    }
  } 
  
/******************************************************************************
 * Play the principal variation from the previous iteration first.            *
 *****************************************************************************/
  
  if (playPVLine && !hasBeenNullMove) {
    for (i=0; i<list->firstFree; i++) {
      if (move_equals(list->moves[i], &(bestLine.argmove[ply]))) {
        tmpMove = list->moves[0];
        list->moves[0] = list->moves[i];
        list->moves[i] = tmpMove;
        nrOfSpecialMoves = 1; 
        break;
      }
    }
  }

/******************************************************************************
 * Put the move probed from the hashtable first in the movelist               *
 *****************************************************************************/
  if (bestMove.promotion != 'F') {
    for (i=0; i<list->firstFree; i++) {
      if (move_equals(list->moves[i], &bestMove)) {
        tmpMove = list->moves[0];
        list->moves[0] = list->moves[i];
        list->moves[i] = tmpMove;
        nrOfSpecialMoves = 1;
        break;
      }
    }
  }
  
/******************************************************************************
 * quick-sort the moves                                                       *
 *****************************************************************************/
  if (list->firstFree > nrOfSpecialMoves + 1 && ply>0) {
    atWhatPlyAreWe = ply;
    helpPosition = &(game->currentPosition);
    qsort(list->moves + nrOfSpecialMoves, list->firstFree-1, sizeof(PMove), 
          cmpMoves);
  }
  
/******************************************************************************
 * Loop through the moves...                                                  *
 * and call this function                                                     *
 *****************************************************************************/
  for (i=0; i<list->firstFree; i++)
  {
    gameCopy = *game;
    makeMove(&gameCopy, list->moves[i]);

    /* Extend pawn pushes to 7th and promotions, then call search */    
    newDepth = depth;
    if (list->moves[i]->piece == 'P' || list->moves[i]->piece == 'p') {
      if (list->moves[i]->y2 == 6 || list->moves[i]->y2 == 1)
        newDepth++;
      else if (list->moves[i]->promotion != NO_PROMOTION)
        newDepth++;
    }
    
    value = -search(&gameCopy, (newDepth - 1), -beta, -alpha, &line, 
                    maxtime, ply+1, maxSearchDepth, FALSE, hasBeenNullMove);

    /* don't go further if we are out of time */
    if (searchInterrupted) {
      freeMoveList(list);
      return 0;
    }

    /* remember the scores of the moves at ply 0 */
    if (ply == 0) {
      moveScoresAtFirstPly.firstFree = i + 1;
      moveScoresAtFirstPly.moveScore[i].move = *(list->moves[i]);
      moveScoresAtFirstPly.moveScore[i].score = value;
    }

    if (value > bestScore) {
      bestScore = value;
      bestMove = *(list->moves[i]);
    }

    /* recordHash, addKillerMove and addHistoryMove when we have a beta 
       cutoff */
    if (value >= beta) {
      recordHash(depth - amount_of_extenssions_done, beta, HASH_BETA, 
                 game->currentPosition.key, list->moves[i]);
      addKillerMove(list->moves[i], ply);
      addHistoryMove(list->moves[i], ply);
      freeMoveList(list);
      return beta;
    }

    /* Our value of this position should be exact, so remember the principal
       variation line. */
    if (value > alpha) {
      alpha = value;
      hashFlag = HASH_EXACT;
      if (!hasBeenNullMove) {
        //printf("a"); fflush(stdout);
        pline->argmove[0] = *(list->moves[i]);
        memcpy(pline->argmove + 1, line.argmove, line.cmove * sizeof(TMove));
        pline->cmove = line.cmove + 1;
        //printf("b"); fflush(stdout);
      }
    }
  }

  /* Store this position also to the hashtable */
  recordHash(depth - amount_of_extenssions_done, alpha, hashFlag, 
             game->currentPosition.key, &bestMove);

  freeMoveList(list);
  
  return alpha;
}


void printLine(TLine *line)
{
  int i; 
  for (i=0; i<line->cmove; i++) {
    p_move(&(line->argmove[i]), stdout);
    printf(" ");
  }
}


int cmpMoveScores(const void *table_entry_ptr1, const void *table_entry_ptr2)
{
  int a = ((TMoveScore *)table_entry_ptr1)->score;
  int b = ((TMoveScore *)table_entry_ptr2)->score;

  return (a == b) ? 0 : ((a > b) ? -1 : 1);
}


#ifdef WINDOWS
/* This function comes from Mikko Nummelin */
double dtime() {
  FILETIME ft;
  double timeindouble;
  GetSystemTimeAsFileTime(&ft);
  timeindouble = (double)ft.dwHighDateTime * 429.4967296 + 
    (double)ft.dwLowDateTime/10000000;
  return timeindouble;
}

#else
/* This function comes from Mikko Nummelin */
double dtime() 
{
  int retval;
  double timeInDouble;
  Timeval tval;

  retval=gettimeofday(&tval,NULL);
  if(retval==-1) 
    return (double)-1;
  timeInDouble=((double)tval.tv_sec+(double)tval.tv_usec/1000000);

  return timeInDouble;
}

#endif

/*****************************************
 * Does a capture search.                *
 ****************************************/
int quiesenceSearch(PGame game, int alpha, int beta)
{
  PMoveList list;
  TGame copy;
  int value, i;

  nodes++;
  value = evaluate(game);

  if (value >= beta)
    return beta;
  if (value > alpha)
    alpha = value;
  list = generateCaptures(&(game->currentPosition));
  if (list->firstFree > 1) {
    helpPosition = &(game->currentPosition);
    qsort(list->moves, list->firstFree-1, sizeof(PMove), fastcmpMoves);
  }

  for (i=0; i<list->firstFree; i++) 
  {
    copy = *game;
    makeMove(&copy, list->moves[i]);      
    value = -quiesenceSearch(&copy, -beta, -alpha);
    if (value >= beta) {
      freeMoveList(list);
      return beta;
    }
    if (value > alpha)
      alpha = value;
  }
  freeMoveList(list);
  return alpha;
}


/*******************************************************************
 * Compares two moves without using history tables or killermoves. *
 * Only MVV/LVA (Most Valuable Victim/Least Valuable Attacker)     *
 ******************************************************************/
int fastcmpMoves(const void *table_entry_ptr1, const void *table_entry_ptr2)
{
  PMove m_a = * (PMove *)table_entry_ptr1;
  PMove m_b = * (PMove *)table_entry_ptr2;
  int a = 0;
  int b = 0;
  int moved, capitured;
  if (m_a->promotion == 'q') a += 500000;
  if (m_b->promotion == 'q') b += 500000;
  if (m_a->capiture) {
    a += 20000;
    moved = getPoints(helpPosition->board[m_a->x1][m_a->y1]);
    capitured = getPoints(helpPosition->board[m_a->x2][m_a->y2]);
    a += (capitured - moved);
  }
  if (m_b->capiture) {    
    b += 20000;
    moved = getPoints(helpPosition->board[m_b->x1][m_b->y1]);
    capitured = getPoints(helpPosition->board[m_b->x2][m_b->y2]);
    b += (capitured - moved);
  }

  return (a == b) ? 0 : ((a > b) ? -1 : 1);
}


/**************************************************************************
 * Compares two moves by using killermoves and history tables and MVV/LVA *
 *************************************************************************/
int cmpMoves(const void *table_entry_ptr1, const void *table_entry_ptr2)
{
  PMove m_a = * (PMove *)table_entry_ptr1;
  PMove m_b = * (PMove *)table_entry_ptr2;
  int a = 0;
  int b = 0;
  int moved, capitured;
  if (m_a->promotion == 'q') a += 50000;
  if (m_b->promotion == 'q') b += 50000;
  if (m_a->capiture) {
    a += 20000;
    moved = getPoints(helpPosition->board[m_a->x1][m_a->y1]);
    capitured = getPoints(helpPosition->board[m_a->x2][m_a->y2]);
    a += (capitured - moved);
  }
  if (m_b->capiture) {    
    b += 20000;
    moved = getPoints(helpPosition->board[m_b->x1][m_b->y1]);
    capitured = getPoints(helpPosition->board[m_b->x2][m_b->y2]);
    b += (capitured - moved);
  }

  /* History move bonus */
  a += getMoveBonus(m_a, atWhatPlyAreWe);
  b += getMoveBonus(m_b, atWhatPlyAreWe);

  // killer moves
  if (move_equals(m_a, killerMoves0+atWhatPlyAreWe))
    a += 9999999-lastKillerMove[atWhatPlyAreWe];
  else if (move_equals(m_b, killerMoves0+atWhatPlyAreWe))
    b += 9999999-lastKillerMove[atWhatPlyAreWe];
  if (move_equals(m_a, killerMoves1+atWhatPlyAreWe))
    a += 9999998+lastKillerMove[atWhatPlyAreWe];
  else if (move_equals(m_b, killerMoves1+atWhatPlyAreWe))
    b += 9999998+lastKillerMove[atWhatPlyAreWe];

  return (a == b) ? 0 : ((a > b) ? -1 : 1);
}


int getPoints(char c)
{
  c = toupper(c);
  if (c == 'Q') return QUEEN_VALUE;
  if (c == 'R') return ROOK_VALUE;
  if (c == 'B') return BISHOP_VALUE;
  if (c == 'N') return KNIGHT_VALUE;
  if (c == 'P') return PAWN_VALUE;
  else return 10000;
}
